/*
 * This file is part of the paper THE PRINCIPLE OF IRREVERSIBILITY
 * IN A STOCHASTIC MODEL OF WARTIME NEGOTIATIONS.
 *
 * Copyright (c) 2002 by Branislav L. Slantchev
 * Dept. of Political Science, University of Rochester
 * Academic email: bsla@troi.cc.rochester.edu
 *
 * This source code is specifically written to produce the simulations
 * that verify that for high discount factor, player 1 will strictly
 * prefer to screen out its opponent up to three periods (this is
 * for the incomplete information case used in the paper). Any other
 * application is at the discretion of the user. This software may
 * be used, copied, redistributed, and modified, as long as changes
 * are not attributed to Branislav L. Slantchev. All warranties
 * are expressly disclaimed. I make no representation about the
 * suitability of this software for any particular purpose.
*/

new;
#include calc-f.src;

/*
** construct the M partition of size n for the A matrix
*/
proc(1)=make_partition_m(n,delta,p);
    local m, left, right, lel, rel;

    lel = delta*(1-p);
    rel = delta*p;
    m = zeros(n,n);
    for i(1,n,1);
        left = i-1;
        right = i+1;
        if( left >= 1 ); m[i,left] = lel; endif;
        if( right <= n); m[i,right] = rel; endif;
    endfor;

    retp(m);
endp;


/*
** construct the matrix of coefficients A for N states of the world
*/
proc(1)=coeff_matrix(N,delta,p);
    local a,i,m;

    m = make_partition_m(N-1,delta,p);    @ make the M partition @
    i = eye(N-1);                        @ make the I partition @
    a = (i~m)|(m~i);                    @ concatenate them @

    retp(a);
endp;

/*
** construct the RHS vector Y for N states of the world
*/
proc(1)=rhs_matrix(N,delta,p,pie,b1,b2);
    local y, first, last, amid, bmid, dim;

    amid = {};
    bmid = {};
    first = pie-delta*(1-p)*pie-(1-delta)*b2;
    last  = pie-delta*p*pie-(1-delta)*b1;
    if(N > 2);
        dim = N-2;
        amid = ones(dim,1)*(pie-(1-delta)*b2);
        bmid = ones(dim,1)*(pie-(1-delta)*b1);
    endif;

    y = first|amid|bmid|last;
    retp(y);
endp;


/*
** compute the optimal offers, x*, from Y and the inverse of A
*/
proc(1)=calc_x(N,delta,p,pie,b1,b2);
    local x,A,Y;

    A = coeff_matrix(N,delta,p);
    Y = rhs_matrix(N,delta,p,pie,b1,b2);
    x = inv(A)*Y;
    retp(x);
endp;

/*
** simple number formatting
*/
proc(1)=std_val(n);
    local aStr;

    aStr=ftos(n,"%#*.*lf",3,2);
    retp(aStr);
endp;

proc calc_bigQ(qS,qM,pL,pM,pH);
	retp( (1-qS-qM).*(qS.*(pH-pL)+qM.*(pH-pM)) );
endp;

/*
 * calculate the messy expression for screening up to three periods
 * if the result is > 0, then screening for three periods is optimal
*/
proc check_diff(delta,pie,b1,b2,x1,k0,qM,qS,pL,pM,pH);
	local a,b,v0,u0,v2,v2m;
	local denom1, denom2,qW;

	a = -qM.*(1-delta).*(pie-b1-b2);
	qW = 1-qS-qM;
	denom1 = 1 / (qS*pL+qM*pM);
	denom2 = 1 / (qS*(1-pL)+qM*(1-pM));

	v2 = qM.*pM.*pM + qW.*pH.*(qS.*pL.*pL+qM.*pM.*pM).*denom1
		 + qS.*pM.*pL - pM.*pL;

	v2m = qM*(1-pM)*(1-pM) + qW*(1-pH)*(qS*(1-pL)*(1-pL)+qM*(1-pM)*(1-pM))*denom2
		  - (1-qS)*(1-pM)*(1-pL);

	v0  = 2*qM*(1-pM)*pM+qW*(qS*(1-pL)*pL+qM*(1-pM)*pM)*(pH*denom1+(1-pH)*denom2)
		  - (1-qS)*(pM+pL-2*pM*pL);
/*
print "in diff: ";
print "------> coeff for v(k+2): " v2;
print "------> coeff for v(k)  : " v0;
print "------> coeff for v(k-2): " v2m;
*/
	b = x1[k0+2,.]*v2 + x1[k0,.]*v0 + x1[k0-2,.]*v2m;

	retp( a + delta*delta*b );
endp;

/*
 * if this is greater than V1s(k0), then we're fine: 1 prefers to screen in 2 periods
*/
proc calc_payoff_screen_two(delta,pie,b1,b2,x2,k0,qS,qM,pL,pM,pH);
	local a,b;

	a = (1-qS).*(pie-(1-delta).*b2 - delta.*(pM.*x2[k0+1,.] + (1-pM).*x2[k0-1,.]) );
	b = qS .* ((1-delta).*b1 + delta.*pie - delta.*(pL.*x2[k0+1,.] + (1-pL).*x2[k0-1,.]) );

	retp(a + b);
endp;

/*
 * check my simplifications
*/
proc screen_2_1(delta,pie,b1,b2,x2,k0,qS,pL,pM);
	local a,b;

	a = pie - ((1-delta)*b2 + delta*(pM*x2[k0+1,.]+(1-pM)*x2[k0-1,.]));
	b = qS*((1-delta)*(pie-b1-b2)+delta*(pM-pL)*(x2[k0-1,.]-x2[k0+1,.]));
	retp( a - b );
endp;

proc screen_2_2(delta,pie,b1,b2,x1,k0,qS,pL,pM);
	local L1, L2, a, b;

	L1 = pL*x1[k0+2,.]+(1-pL)*x1[k0,.];
	L2 = pL*x1[k0,.] + (1-pL)*x1[k0-2,.];
	a  = (1-qS)*(1-delta)*(pie-b1-b2)+(1-delta*delta)*b1;
	b  = (1-qS)*(pM*L1+(1-pM)*L2) + qS*(pL*L1+(1-pL)*L2);

	retp( a + delta*delta*b );
endp;

/*
 * delta should be less than the result of this function!
*/
proc min_delta_2_periods(delta,pie,b1,b2,x2,k0,qS,pL,pM);
	local a, b;

	a = (1-delta) .* (qS ./ (1-qS));
	b = (pie-b1-b2) ./ ((pM-pL).*(x2[k0-1,.]-x2[k0+1,.]));
	retp( a .* b );
endp;

/*
 * calculate updated beliefs qS(1) and qS(0)
*/
proc qS_victory(qS,qM,pL,pM);
	retp( (qS.*pL) ./ (qS.*pL+qM.*pM) );
endp;

proc qS_defeat(qS,qM,pL,pM);
	retp( (qS.*(1-pL)) ./ (qS.*(1-pL)+qM.*(1-pM)) );
endp;

/*
 * minimum delta required for player 2 make an unacceptable offer in the
 * second period. Note that we need to pass qS1 and k1, not qS and k0
*/
proc min_delta_for_unacceptable(delta,pie,b1,b2,x1,k1,qS1,pL,pM);
	local a,b;

	a = (1-delta).*(pie-b1-b2);
	b = (1-qS1).*(pM-pL).*(x1[k1+1,.]-x1[k1-1,.]);
	retp(a ./ b);
endp;

/*
 * calculate x^*(q,k)
*/
proc x_star(delta,pie,b1,b2,x1,k0,qS,qM,pL,pM,pH);
	local v2,v0,v2m,a,b,rv;

	v2 = x1[k0+2,.]*(qS*pL*pL+qM*pM*pM);
	v0 = x1[k0,.]*(qS*(1-pL)*pL+qM*(1-pM)*pM);
	v2m= x1[k0-2,.]*(qS*(1-pL)*(1-pL)+qM*(1-pM)*(1-pM));
	a  = 1 / (qS*pL + qM*pM);
	b  = 1 / (qS*(1-pL)+qM*(1-pM));

	rv = (1-delta)*(pie-b2+delta*b1)+delta*delta*(pH*a*v2
			+(pH*a+(1-pH)*b)*v0 + (1-pH)*b*v2m );

	retp( rv );
endp;

/*
 * calculate second part of RHS (the sum)
*/
proc rhs_sum(delta,b1,x1,k0,qS,qM,pL,pM);
	local v2,v0,v2m,c,rv;

	v2 = x1[k0+2,.]*(qS*pL*pL+qM*pM*pM);
	v0 = x1[k0,.]*(qS*(1-pL)*pL+qM*(1-pM)*pM);
	v2m= x1[k0-2,.]*(qS*(1-pL)*(1-pL)+qM*(1-pM)*(1-pM));
	c  = 1 / (qS+qM);

	rv = (1-delta)*b1+delta*c*(v2+2*v0+v2m);

	retp(rv);
endp;

/*
 * calculate the expected payoff from screening for three periods
*/
proc rhs_screen_3(delta,pie,b1,b2,x1,k0,qS,qM,pL,pM,pH);
	local x,s,rv,qW;

	x = x_star(delta,pie,b1,b2,x1,k0,qS,qM,pL,pM,pH);
	s = rhs_sum(delta,b1,x1,k0,qS,qM,pL,pM);
	qW = 1-qS-qM;

	rv = qW*x + (1-qW)*((1-delta)*b1+delta*s);

	retp(rv);
endp;

/*
 * check the calculation above!
*/
proc rhs_check(delta,pie,b1,b2,x1,k0,qS,qM,pL,pM,pH);
	local v2,v0,v2m,a,b,c,qW,rv;

	v2 = x1[k0+2,.]*(qS*pL*pL+qM*pM*pM);
	v0 = x1[k0,.]*(qS*(1-pL)*pL+qM*(1-pM)*pM);
	v2m= x1[k0-2,.]*(qS*(1-pL)*(1-pL)+qM*(1-pM)*(1-pM));
	a  = 1 / (qS*pL + qM*pM);
	b  = 1 / (qS*(1-pL)+qM*(1-pM));
	c  = 1 / (qS+qM);
	qW = 1-qS-qM;
	
	rv = qW*(1-delta)*(pie-b1-b2)+(1-delta*delta)*b1 
		+delta*delta*( v2*(1+qW*pH*a) + v0*(2+qW*(pH*a+(1-pH)*b))
			+ v2m*(1+qW*(1-pH)*b) );

	retp(rv);
endp;

/*
 * calculate player 2's optimal proposal in t=1: y^*(qS1(),k1)
*/
proc y_star_victory(delta,pie,b1,x1,k0,qSv,pL,pM);
	local rv,p1,k1;

    p1 = qSv*pL+(1-qSv)*pM;
    k1 = k0+1;

    rv = pie - (1-delta)*b1 - delta*(p1*x1[k1+1,.]+(1-p1)*x1[k1-1,.]);
	retp(rv);
endp;

proc y_star_defeat(delta,pie,b1,x1,k0,qSd,pL,pM);
	local rv,p1,k1;

    p1 = qSd*pL+(1-qSd)*pM;
    k1 = k0-1;

    rv = pie - (1-delta)*b1 - delta*(p1*x1[k1+1,.]+(1-p1)*x1[k1-1,.]);
	retp(rv);
endp;

/*
 * maximum proposal 2m will accept in initial period: 
 * is smaller than x^*(q,k0), then 2m (and 2s) reject the offer
*/
proc max_2m_t0(delta,pie,b1,b2,x1,k0,qS,qM,pL,pM);
	local rv,y1,y0,qSd,qSv;

	qSv = qS_victory(qS,qM,pL,pM);
	qSd = qS_defeat(qS,qM,pL,pM);
	y1 = y_star_victory(delta,pie,b1,x1,k0,qSv,pL,pM);
	y0 = y_star_defeat(delta,pie,b1,x1,k0,qSd,pL,pM);
	rv = pie-(1-delta)*b2 - delta*(pM*y1 + (1-pM)*y0);
	retp( rv );
endp;

/*
 * same for 2s: checks if it is better to offer y* in t=1
 * or wait for x=V1s in period t=2
*/
proc max_2s_t0(delta,pie,b1,b2,x1,k0,qS,qM,pL,pM);
	local xt1,xt2,xmax,y1,y0,qSd,qSv;

	qSv = qS_victory(qS,qM,pL,pM);
	qSd = qS_defeat(qS,qM,pL,pM);
	y1 = y_star_victory(delta,pie,b1,x1,k0,qSv,pL,pM);
	y0 = y_star_defeat(delta,pie,b1,x1,k0,qSd,pL,pM);
	xt1 = pie-(1-delta)*b2 - delta*(pL*y1 + (1-pL)*y0);
	@ assumes the game won't end in three periods @
	xt2 = pie- (1-delta*delta)*b2 - delta*delta
		*( pL*pL*(pie-x1[k0+2,.]) + (1-pL)*(1-pL)*(pie-x1[k0-2,.])
		  + 2*pL*(1-pL)*(pie-x1[k0,.]) );

    print ">>>> Ex ante 2s pay if offer same as 2m in t=1: " pie-xt1;
    print ">>>> Ex ante 2s pay if nonserious offer in t=1: " pie-xt2;

	if( xt1 > xt2 ); xmax = xt2; else; xmax = xt1; endif;
	retp( xmax );
endp;

/*
 * expected payoff for 2s in t=1 if player 1 rejects an offer
*/
proc pay_2s_t1(delta,pie,b2,x1,k1,pL);
    local v;

    v = (1-delta)*b2 + delta*(pL*(pie-x1[k1+1,.]) + (1-pL)*(pie-x1[k1-1,.]));
    retp(v);
endp;

/*
 * expected payoff for 2m in t=1 if player 1 rejects an offer
*/
proc pay_2m_t1(delta,pie,b2,x1,k1,pM);
    local v;

    v = (1-delta)*b2 + delta*(pM*(pie-x1[k1+1,.]) + (1-pM)*(pie-x1[k1-1,.]));
    retp(v);
endp;


/*****************************************************************
** THE MAIN ROUTINE
******************************************************************/
cls;

N=10;           @ N >= 2 (number of contestable objects) @
delta=.95;      @ common discount factor: .8 and .9 for pooling/partially-separating @
b1=.30;         @ per-period payoff for player 1 @
b2=.40;         @ per-period payoff for player 2 @
pie=1;          @ size of the flow of benefits to divide @
k0 = 6;			@ starting state to test for @
qS = .2;		@ prior that 2 is strong @
qM = .5;		@ prior that 2 is moderately strong @
pL = .4;		@ prob of winning against strong 2 @
pM = .65;		@ prob of winning against moderate 2 @
pH = .75;		@ prob of winning against weak 2 @

x = calc_x(N,delta,pL,pie,b1,b2);		@ calc V^1_s(k) @
k  = seqa(1,1,(N-1));
x1 = x[1:(N-1),.];
x2 = x[N:2*(N-1),.];
f1 = calc_F1(N,pL,delta,b1,pie);
f2 = calc_F2(N,pL,delta,b2,pie);

format /mat /on /ma1 /rd 8,4;
print "\nRunning example against strong opponent (pL):\n";
print "   N     = " N;
print "   delta = " delta;
print "   pL    = " pL;
print "   pie   = " pie;
print "   b1    = " b1;
print "   b2    = " b2;
print "--------------------------------------------------------------";
print "    k      V1s(k)   pie-V1s   x2(k)    pie-x2  W1s(k)   W2s(k) ";
print "--------------------------------------------------------------";
print k~x1~(pie-x1)~x2~(pie-x2)~f1~f2;

if( (x1 < f1) or ((pie-x1) < f2) );
	print "Something's terribly wrong with x1";
endif;

if( (x2 < f2) or ((pie-x2) < f1) );
	print "Something's terribly wrong with x2";
endif;

pay_screen2 = calc_payoff_screen_two(delta,pie,b1,b2,x2,k0,qS,qM,pL,pM,pH);
	@check1 = screen_2_1(delta,pie,b1,b2,x2,k0,qS,pL,pM);@
	@check2 = screen_2_2(delta,pie,b1,b2,x1,k0,qS,pL,pM);@

pay3 = rhs_screen_3(delta,pie,b1,b2,x1,k0,qS,qM,pL,pM,pH);
	@check3 = rhs_check(delta,pie,b1,b2,x1,k0,qS,qM,pL,pM,pH);@
	@print "pay3 = " pay3 "; and check3 = " check3;@

min_delta = min_delta_2_periods(delta,pie,b1,b2,x2,k0,qS,pL,pM);
qS1_victory = qS_victory(qS,qM,pL,pM);
qS1_defeat  = qS_defeat(qS,qM,pL,pM);
min_delta_1_victory = min_delta_for_unacceptable(delta,pie,b1,b2,x1,k0+1,qS1_victory,pL,pM);
min_delta_1_defeat  = min_delta_for_unacceptable(delta,pie,b1,b2,x1,k0-1,qS1_defeat,pL,pM);

print "Player 1's payoff if it offers V1s(k0) in t=0 : " x1[k0,.];
print "Player 1's payoff from separating 2s only     : " pay_screen2;
print "Player 1's payoff from screening in 3 periods : " pay3;

diff = check_diff(delta,pie,b1,b2,x1,k0,qM,qS,pL,pM,pH);
if( diff <= 0 );
	print ">>>> Screening for 3 periods is not optimal; diff : " diff;
else;
	print ">>>> Screening for 3 periods is optimal; diff : " diff;
endif;

qS1v = qS_victory(qS,qM,pL,pM);
qS1d = qS_defeat(qS,qM,pL,pM);
print "Player 1's belief following victory, qS1(1)   : " qS1v;
print "Player 1's belief following defeat, qS1(0)    : " qS1d;

print "Minimum discount factor for 2-period screening: " min_delta;
print "Min delta for unacceptable offer after victory: " min_delta_1_victory;
print "Min delta for unacceptable offer after defeat : " min_delta_1_defeat;

init_offer = x_star(delta,pie,b1,b2,x1,k0,qS,qM,pL,pM,pH);
y_victory = y_star_victory(delta,pie,b1,x1,k0,qS1v,pL,pM);
y_defeat  = y_star_defeat(delta,pie,b1,x1,k0,qS1d,pL,pM);
max_accept_2m = max_2m_t0(delta,pie,b1,b2,x1,k0,qS,qM,pL,pM);
max_accept_2s = max_2s_t0(delta,pie,b1,b2,x1,k0,qS,qM,pL,pM);
pay_2m_victory = pay_2m_t1(delta,pie,b2,x1,k0+1,pM);
pay_2m_defeat = pay_2m_t1(delta,pie,b2,x1,k0-1,pM);
pay_2s_victory = pay_2s_t1(delta,pie,b2,x1,k0+1,pL);
pay_2s_defeat = pay_2s_t1(delta,pie,b2,x1,k0-1,pL);

print "x^*(qS,k0)    : " init_offer " <-- player 1's initial offer";
print "2m's accept   : " max_accept_2m " <-- max x that 2m accepts in t=0";
print "2s's accept   : " max_accept_2s " <-- max x that 2s accepts in t=0";
print "y*(qS1v,k0+1) : " y_victory  " <-- max y player 1 accepts in t=1 after victory for 1";
print "y*(qS1d,k0-1) : " y_defeat   " <-- max y player 1 accepts in t=1 after defeat of 1";
print "2m pay (k0+1) : " pay_2m_victory " <-- 2m expected pay if 1 rejects offer in t=1 after victory for 1";
print "2m pay (k0-1) : " pay_2m_defeat " <-- 2m expected pay if 1 rejects offer in t=1 after defeat of 1";
print "2s pay (k0+1) : " pay_2s_victory " <-- 2s expected pay if 1 rejects offer in t=1 after victory for 1";
print "2s pay (k0-1) : " pay_2s_defeat " <-- 2s expected pay if 1 rejects offer in t=1 after defeat of 1";

if( (pay_2m_victory < y_victory) and (pay_2m_defeat < y_defeat) );
    print ">>>> Player 2m prefers to offer y* in t=1";
else;
    print ">>>> Something's wrong: 2m prefers to make a nonserious offer in t=1";
endif;

if( (pay_2s_victory > y_victory) and (pay_2s_defeat > y_defeat) );
    print ">>>> Player 2s prefers to make a nonserious offer in t=1";
else;
    print ">>>> Something's wrong: 2s prefers to offer whatever 2m offers in t=1";
endif;

/*
 * we now have to compute the payoffs for player 2w and 2m
*/
@ recompute fighting to the end @
f1s = f1;
f1m = calc_F1(N,pM,delta,b1,pie);
f1w = calc_F1(N,pH,delta,b1,pie);

f2s = f2;
f2m = calc_F2(N,pM,delta,b2,pie);
f2w = calc_F2(N,pH,delta,b2,pie);

print "---------------------------------------------------------";
print "Player 1's payoffs from fighting 2s 2m 2w";
print "---------------------------------------------------------";
print k~f1s~f1m~f1w;

print "---------------------------------------------------------";
print "Player 2s 2m 2w payoffs from fighting to the end";
print "---------------------------------------------------------";
print k~f2s~f2m~f2w;

xw = calc_x(N,delta,pH,pie,b1,b2);		@ calc V^2_w(k) @
x1w = xw[1:(N-1),.];
x2w = xw[N:2*(N-1),.];
print "-----------------------------------------------";
print "Optimal offers if playing 2w with complete info";
print "   k       x1(k)    pie-x1   x2(k)    pie-x2   ";
print "-----------------------------------------------";
print k~x1w~(pie-x1w)~x2w~(pie-x2w);

xm = calc_x(N,delta,pM,pie,b1,b2);		@ calc V^2_m(k) @
x1m = xm[1:(N-1),.];
x2m = xm[N:2*(N-1),.];
print "-----------------------------------------------";
print "Optimal offers if playing 2m with complete info";
print "   k       x1(k)    pie-x1   x2(k)    pie-x2   ";
print "-----------------------------------------------";
print k~x1m~(pie-x1m)~x2m~(pie-x2m);
